# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gui.ui',
# licensing of 'gui.ui' applies.
#
# Created: Tue Aug 20 18:54:58 2019
#      by: pyside2-uic  running on PySide2 5.13.0
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(1183, 675)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(Dialog)
        self.tabWidget.setObjectName("tabWidget")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.tab_2)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.main_splitter = QtWidgets.QSplitter(self.tab_2)
        self.main_splitter.setOrientation(QtCore.Qt.Horizontal)
        self.main_splitter.setObjectName("main_splitter")
        self.frame_8 = QtWidgets.QFrame(self.main_splitter)
        self.frame_8.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_8.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_8.setObjectName("frame_8")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.frame_8)
        self.verticalLayout_5.setContentsMargins(1, 1, 1, 1)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.frame_5 = QtWidgets.QFrame(self.frame_8)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_5.sizePolicy().hasHeightForWidth())
        self.frame_5.setSizePolicy(sizePolicy)
        self.frame_5.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_5.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_5.setObjectName("frame_5")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame_5)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_9 = QtWidgets.QLabel(self.frame_5)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout.addWidget(self.label_9)
        self.name_lineEdit = QtWidgets.QLineEdit(self.frame_5)
        self.name_lineEdit.setObjectName("name_lineEdit")
        self.horizontalLayout.addWidget(self.name_lineEdit)
        self.verticalLayout_5.addWidget(self.frame_5)
        self.frame_6 = QtWidgets.QFrame(self.frame_8)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_6.sizePolicy().hasHeightForWidth())
        self.frame_6.setSizePolicy(sizePolicy)
        self.frame_6.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_6.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_6.setObjectName("frame_6")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.frame_6)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.label_8 = QtWidgets.QLabel(self.frame_6)
        self.label_8.setObjectName("label_8")
        self.horizontalLayout_3.addWidget(self.label_8)
        self.frequency_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.frame_6)
        self.frequency_doubleSpinBox.setDecimals(0)
        self.frequency_doubleSpinBox.setProperty("value", 50.0)
        self.frequency_doubleSpinBox.setObjectName("frequency_doubleSpinBox")
        self.horizontalLayout_3.addWidget(self.frequency_doubleSpinBox)
        self.label_11 = QtWidgets.QLabel(self.frame_6)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout_3.addWidget(self.label_11)
        self.rho_doubleSpinBox = QtWidgets.QDoubleSpinBox(self.frame_6)
        self.rho_doubleSpinBox.setMaximum(9999999.0)
        self.rho_doubleSpinBox.setProperty("value", 100.0)
        self.rho_doubleSpinBox.setObjectName("rho_doubleSpinBox")
        self.horizontalLayout_3.addWidget(self.rho_doubleSpinBox)
        self.verticalLayout_5.addWidget(self.frame_6)
        self.frame_2 = QtWidgets.QFrame(self.frame_8)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_2.sizePolicy().hasHeightForWidth())
        self.frame_2.setSizePolicy(sizePolicy)
        self.frame_2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.frame_2)
        self.verticalLayout_9.setContentsMargins(1, 1, 1, 1)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.splitter = QtWidgets.QSplitter(self.frame_2)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.frame_3 = QtWidgets.QFrame(self.splitter)
        self.frame_3.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_3.setObjectName("frame_3")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.frame_3)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.label_12 = QtWidgets.QLabel(self.frame_3)
        self.label_12.setObjectName("label_12")
        self.verticalLayout_8.addWidget(self.label_12)
        self.wires_tableView = QtWidgets.QTableView(self.frame_3)
        self.wires_tableView.setObjectName("wires_tableView")
        self.verticalLayout_8.addWidget(self.wires_tableView)
        self.frame_4 = QtWidgets.QFrame(self.splitter)
        self.frame_4.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_4.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_4.setObjectName("frame_4")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.frame_4)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_10 = QtWidgets.QLabel(self.frame_4)
        self.label_10.setObjectName("label_10")
        self.verticalLayout_4.addWidget(self.label_10)
        self.tower_tableView = QtWidgets.QTableView(self.frame_4)
        self.tower_tableView.setObjectName("tower_tableView")
        self.verticalLayout_4.addWidget(self.tower_tableView)
        self.frame = QtWidgets.QFrame(self.frame_4)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.frame)
        self.horizontalLayout_2.setContentsMargins(0, -1, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.add_to_tower_pushButton = QtWidgets.QPushButton(self.frame)
        self.add_to_tower_pushButton.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/Icons/icons/plus.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.add_to_tower_pushButton.setIcon(icon)
        self.add_to_tower_pushButton.setObjectName("add_to_tower_pushButton")
        self.horizontalLayout_2.addWidget(self.add_to_tower_pushButton)
        self.delete_from_tower_pushButton = QtWidgets.QPushButton(self.frame)
        self.delete_from_tower_pushButton.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/Icons/icons/minus.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.delete_from_tower_pushButton.setIcon(icon1)
        self.delete_from_tower_pushButton.setObjectName("delete_from_tower_pushButton")
        self.horizontalLayout_2.addWidget(self.delete_from_tower_pushButton)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.compute_pushButton = QtWidgets.QPushButton(self.frame)
        self.compute_pushButton.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/Icons/icons/calc.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.compute_pushButton.setIcon(icon2)
        self.compute_pushButton.setIconSize(QtCore.QSize(32, 32))
        self.compute_pushButton.setObjectName("compute_pushButton")
        self.horizontalLayout_2.addWidget(self.compute_pushButton)
        self.verticalLayout_4.addWidget(self.frame)
        self.verticalLayout_9.addWidget(self.splitter)
        self.verticalLayout_5.addWidget(self.frame_2)
        self.PlotFrame = QtWidgets.QFrame(self.main_splitter)
        self.PlotFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.PlotFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.PlotFrame.setObjectName("PlotFrame")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.PlotFrame)
        self.verticalLayout_7.setContentsMargins(9, 9, 9, 9)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.label_4 = QtWidgets.QLabel(self.PlotFrame)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_7.addWidget(self.label_4)
        self.plotwidget = MatplotlibWidget(self.PlotFrame)
        self.plotwidget.setObjectName("plotwidget")
        self.verticalLayout_7.addWidget(self.plotwidget)
        self.verticalLayout_6.addWidget(self.main_splitter)
        self.tabWidget.addTab(self.tab_2, "")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tab)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label = QtWidgets.QLabel(self.tab)
        self.label.setObjectName("label")
        self.verticalLayout_3.addWidget(self.label)
        self.z_tableView_abcn = QtWidgets.QTableView(self.tab)
        self.z_tableView_abcn.setObjectName("z_tableView_abcn")
        self.verticalLayout_3.addWidget(self.z_tableView_abcn)
        self.label_6 = QtWidgets.QLabel(self.tab)
        self.label_6.setObjectName("label_6")
        self.verticalLayout_3.addWidget(self.label_6)
        self.z_tableView_abc = QtWidgets.QTableView(self.tab)
        self.z_tableView_abc.setObjectName("z_tableView_abc")
        self.verticalLayout_3.addWidget(self.z_tableView_abc)
        self.label_7 = QtWidgets.QLabel(self.tab)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_3.addWidget(self.label_7)
        self.z_tableView_seq = QtWidgets.QTableView(self.tab)
        self.z_tableView_seq.setObjectName("z_tableView_seq")
        self.verticalLayout_3.addWidget(self.z_tableView_seq)
        self.tabWidget.addTab(self.tab, "")
        self.tab_3 = QtWidgets.QWidget()
        self.tab_3.setObjectName("tab_3")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tab_3)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_2 = QtWidgets.QLabel(self.tab_3)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.y_tableView_abcn = QtWidgets.QTableView(self.tab_3)
        self.y_tableView_abcn.setObjectName("y_tableView_abcn")
        self.verticalLayout_2.addWidget(self.y_tableView_abcn)
        self.label_3 = QtWidgets.QLabel(self.tab_3)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_2.addWidget(self.label_3)
        self.y_tableView_abc = QtWidgets.QTableView(self.tab_3)
        self.y_tableView_abc.setObjectName("y_tableView_abc")
        self.verticalLayout_2.addWidget(self.y_tableView_abc)
        self.label_5 = QtWidgets.QLabel(self.tab_3)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_2.addWidget(self.label_5)
        self.y_tableView_seq = QtWidgets.QTableView(self.tab_3)
        self.y_tableView_seq.setObjectName("y_tableView_seq")
        self.verticalLayout_2.addWidget(self.y_tableView_seq)
        self.tabWidget.addTab(self.tab_3, "")
        self.verticalLayout.addWidget(self.tabWidget)

        self.retranslateUi(Dialog)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QtWidgets.QApplication.translate("Dialog", "Tower creation", None, -1))
        self.label_9.setText(QtWidgets.QApplication.translate("Dialog", "Name", None, -1))
        self.label_8.setText(QtWidgets.QApplication.translate("Dialog", "Frequency (Hz)", None, -1))
        self.label_11.setText(QtWidgets.QApplication.translate("Dialog", "Earth resistivity (Ohm/m^3)", None, -1))
        self.label_12.setText(QtWidgets.QApplication.translate("Dialog", "Wire catalogue", None, -1))
        self.label_10.setText(QtWidgets.QApplication.translate("Dialog", "Wire compisition", None, -1))
        self.add_to_tower_pushButton.setToolTip(QtWidgets.QApplication.translate("Dialog", "Add wire", None, -1))
        self.delete_from_tower_pushButton.setToolTip(QtWidgets.QApplication.translate("Dialog", "Delete wire", None, -1))
        self.compute_pushButton.setToolTip(QtWidgets.QApplication.translate("Dialog", "Compute matrices", None, -1))
        self.label_4.setText(QtWidgets.QApplication.translate("Dialog", "Tower", None, -1))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), QtWidgets.QApplication.translate("Dialog", "Tower", None, -1))
        self.label.setText(QtWidgets.QApplication.translate("Dialog", "   Z series (Ohm / km) for ABCN", None, -1))
        self.label_6.setText(QtWidgets.QApplication.translate("Dialog", "   Z series (Ohm / km) for ABC", None, -1))
        self.label_7.setText(QtWidgets.QApplication.translate("Dialog", "   Z series (Ohm / km) in sequence components", None, -1))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), QtWidgets.QApplication.translate("Dialog", "Z series", None, -1))
        self.label_2.setText(QtWidgets.QApplication.translate("Dialog", "   Y shunt (uS / km) for ABCN", None, -1))
        self.label_3.setText(QtWidgets.QApplication.translate("Dialog", "   Y shunt (uS / km) for ABC", None, -1))
        self.label_5.setText(QtWidgets.QApplication.translate("Dialog", "   Y shunt (uS / km) for the sequence components", None, -1))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_3), QtWidgets.QApplication.translate("Dialog", "Y shunt", None, -1))

from .matplotlibwidget import MatplotlibWidget
from .icons_rc import *

if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())

